package com.flexdms.jsfutils;

import java.beans.FeatureDescriptor;
import java.util.Iterator;
import java.util.logging.Logger;

import javax.el.ELContext;
import javax.el.ELResolver;
import javax.el.PropertyNotWritableException;
import javax.faces.component.NamingContainer;
import javax.faces.component.UIComponent;
import javax.faces.component.UIData;
import javax.faces.context.FacesContext;

/**
 * Resolve property by search component attributes. It will search the component tree from current component to UIViewRoot unless one is found.
 * @author jason
 * 
 * _cdattrs: search NamingContainer including UIData
 * _ccattrs: search namingContaining excluding UIData
 * _cattrs: seacrh every UIComponent.
 * 
 * If used as base, find from current component.
 * If used as property name, base should be UIComponent. Find from that component.
 * 
 * 
 * Usage:
 * #{_ccattrs.prop}/#{_cdattrs.prop}/#{_cattrs.prop} search from current component, 
 * 		1. search 
 * 			_ccattrs:NamingContainer excluding UIData
 * 			_cdattrs:NamingContainer including UIData
 * 			_cattrs:Component.
 * 			Set the search component as "A"
 * 		2. call A.getAttributes().get(prop), If it does not return null, return it
 * 			if A is UIData, also check request attributes which is exposed by UIData
 * 			(this is a hack)
 * 			If it return null, set current component to A.parent, go back step 1.
 *  #{A._ccattrs.prop}/#{A._cdattrs.prop}/#{A.cattrs.prop}
 *  	Instead of searching from current component, search from component A. A should be resolved to UIComponent.
 *   
 */
public class AttrsPropertyResolver extends ELResolver
{
	
	public static class ComponentContext
	{
		AttrsHolder holder;
		UIComponent component;
		public AttrsHolder getHolder()
		{
			return holder;
		}
		public void setHolder(AttrsHolder holder)
		{
			this.holder = holder;
		}
		public UIComponent getComponent()
		{
			return component;
		}
		public void setComponent(UIComponent component)
		{
			this.component = component;
		}
		public ComponentContext(AttrsHolder holder, UIComponent component)
		{
			super();
			this.holder = holder;
			this.component = component;
		}
		
	}
	
	public static enum AttrsHolder
	{
		CDATTRS("_cdattrs"), CCATTRS("_ccattrs"), CATTRS("_cattrs");
		private String symbol;

		private AttrsHolder(String symbol)
		{
			this.symbol = symbol;
		}

		public String getSymbol()
		{
			return symbol;
		}

		public void setSymbol(String symbol)
		{
			this.symbol = symbol;
		}

		public static AttrsHolder toHolder(String symbol)
		{
			for (AttrsHolder ah : AttrsHolder.values())
			{
				if (ah.getSymbol().equals(symbol))
				{
					return ah;
				}
			}
			return null;
		}
	};
	
	final static Logger logger = Logger.getLogger(AttrsPropertyResolver.class
			.getName());
	
	/**
	 * Attach property as attribute to Naming Container including UIData
	 */
	public static final String CDATTRS=AttrsHolder.CDATTRS.getSymbol();
	
	/**
	 * Attach property as attribute to Naming Container except UIData
	 */
	public static final String CCATTRS=AttrsHolder.CCATTRS.getSymbol();
	/**
	 * Attach property as Attribute to Component;
	 */
	public static final String CATTRS=AttrsHolder.CATTRS.getSymbol();

	public static final boolean shouldHandle(Object base, Object property)
	{
		if (base==null)
		{
			if (property!=null && property instanceof String)
			{
				String prop_key=(String)property;
				AttrsHolder ah=AttrsHolder.toHolder(prop_key);
				if (ah!=null)
				{
					return true;
				}
			}
		} else
		{
			if (base instanceof AttrsHolder && property!=null && property instanceof String)
			{
				return true;
			} else if (base instanceof UIComponent && property!=null && property instanceof String)
			{
				String prop_key=(String)property;
				AttrsHolder ah=AttrsHolder.toHolder(prop_key);
				if (ah!=null)
				{
					return true;
				}	
			} else if (base instanceof ComponentContext && property!=null && property instanceof String )
			{
				return true;
			}
		}
		return false;
	}


	protected AttrsHolder resolveBase(Object property)
	{
		String prop_key=(String)property;
		return AttrsHolder.toHolder(prop_key);
	}
	
	
	
	protected Object resolveProperty(Object base, Object property)
	{
		if (base instanceof AttrsHolder)
		{
			AttrsHolder ah=(AttrsHolder)base;
			String prop_key=(String)property;
			return resolveAttributes(ah, prop_key, null);
		} else if (base instanceof UIComponent)
		{
			
			UIComponent component=(UIComponent)base;
			String prop_key=(String)property;
			AttrsHolder ah=AttrsHolder.toHolder(prop_key);
			return  new ComponentContext(ah, component);
		} else if (base instanceof ComponentContext)
		{
			ComponentContext ctx=(ComponentContext)base;
			return resolveAttributes(ctx.getHolder(), (String)property, ctx.getComponent());
		}
		return null;
	}
	
	private Object resolveAttributes(AttrsHolder ah, String prop_key, UIComponent comp)
	{
		
		UIComponent component=Utils.getInViewComponent(comp, FacesContext.getCurrentInstance());
		switch (ah)
		{
		case CATTRS:
			
			while (component != null)
			{
				Object prop = component.getAttributes().get(prop_key);
				if (prop != null)
					return prop;
				component = component.getParent();
			}
			break;
		case CDATTRS:
			while (component != null)
			{
				if (component instanceof NamingContainer)
				{
					Object prop = component.getAttributes().get(prop_key);
					if (prop != null)
						return prop;
					
					if (component instanceof UIData)
					{
						UIData data=(UIData)component;
						if (data.getVar().equals(prop_key))
						{
							//TODO test the data.getVar logic
							prop=FacesContext.getCurrentInstance()
									.getExternalContext().getRequestMap().get(prop_key);
						
							if (prop!=null)
							{
								return prop;
							}
						}
					}
				}
				component = component.getParent();
			}
			break;
		case CCATTRS:
			while (component != null)
			{
				if ((component instanceof NamingContainer)
						&& !(component instanceof UIData))
				{
					Object prop = component.getAttributes().get(prop_key);
					if (prop != null)
						return prop;
				}
				component = component.getParent();
			}
			break;
		}
		return null;
	}
	
	@Override
	public Object getValue(ELContext context, Object base, Object property)
	{
		if (shouldHandle(base, property))
		{
			context.setPropertyResolved(true);
			if (base==null)
			{
				return resolveBase(property);
			} else
			{
				Object obj=resolveProperty(base, property);
				if (obj==null)
				{
					logger.warning("can not resolve "+property+ "by AttrsPropetryResolver");
				}
				return obj;
			}
		}
		
		return null;
			
			
	}

	@Override
	public Class<?> getType(ELContext context, Object base, Object property)
	{
		if (shouldHandle(base, property))
		{
			/*
			 * No object is accepted.
			 */
			context.setPropertyResolved(true);
			if (base==null)
			{
				return AttrsHolder.class;
			} else
			{
				return Object.class;
			}
		}
		return null;
	}

	@Override
	public void setValue(ELContext context, Object base, Object property,
			Object value)
	{
		if (shouldHandle(base, property))
		{
			throw new PropertyNotWritableException();
		}

	}

	@Override
	public boolean isReadOnly(ELContext context, Object base, Object property)
	{
		if (shouldHandle(base, property))
		{
			context.setPropertyResolved(true);
			return true;
		}
		return false;
	}

	@Override
	public Iterator<FeatureDescriptor> getFeatureDescriptors(ELContext context,
			Object base)
	{
		return null;
	}

	@Override
	public Class<?> getCommonPropertyType(ELContext context, Object base)
	{
		if (base == null || base instanceof AttrsHolder)
		{
			return String.class;
		}
		return null;
	}

}
